// board.cc
// Ian Lance Taylor for ICFP 2002

// Implementation of Board, which represents the game board.

#include <cassert>
#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <map>
#include <algorithm>

#include "debug.h"
#include "socket.h"
#include "board.h"

void
in_direction(
  Direction d,
  int& x,
  int& y)
{
  switch (d)
    {
    case direction_north:
      ++y;
      break;
    case direction_south:
      --y;
      break;
    case direction_east:
      ++x;
      break;
    case direction_west:
      --x;
      break;
    default:
      assert(0);
    }
}

Direction
direction_for_char(
  char d)
{
  switch (d)
    {
    case 'N':
      return direction_north;
    case 'S':
      return direction_south;
    case 'E':
      return direction_east;
    case 'W':
      return direction_west;
    default:
      assert(0);
    }
}

char
char_for_direction(
  Direction d)
{
  switch (d)
    {
    case direction_north:
      return 'N';
    case direction_south:
      return 'S';
    case direction_east:
      return 'E';
    case direction_west:
      return 'W';
    default:
      assert(0);
    }
}

Direction
reverse_direction(
  Direction d)
{
  switch (d)
    {
    case direction_north:
      return direction_south;
    case direction_south:
      return direction_north;
    case direction_east:
      return direction_west;
    case direction_west:
      return direction_east;
    default:
      assert(0);
    }
}

Direction
direction_right(
  Direction d)
{
  switch (d)
    {
    case direction_north:
      return direction_east;
    case direction_south:
      return direction_west;
    case direction_east:
      return direction_south;
    case direction_west:
      return direction_north;
    default:
      assert(0);
    }
}

std::ostream&
operator<<(
  std::ostream& os,
  Direction d)
{
  os << char_for_direction(d);
  return os;
}

bool
Distances::known(
  int x,
  int y) const
{
  return (this->m_distances.find(Xy(x, y))
	  != this->m_distances.end());
}

int
Distances::to(
  int x,
  int y) const
{
  Distance_map::const_iterator p;
  p = this->m_distances.find(Xy(x, y));
  if (p == this->m_distances.end())
    { return -1; }
  return p->second;
}

void
Board::read(
  Socket& sock)
{
  std::string dimensions = sock.getline();
  char* end;
  this->m_width = strtol(dimensions.c_str(), &end, 10);
  if (*end != ' ')
    {
      std::clog << "syntax error in board dimensions: " << dimensions
		<< std::endl;
      exit(1);
    }

  this->m_height = strtol(end, 0, 10);

  if ((this->m_width < 1 || this->m_width > 10000)
      || (this->m_height < 1 || this->m_height > 10000))
    {
      std::clog << "bad board dimensions: " << this->m_width << " "
		<< this->m_height << std::endl;
      exit(1);
    }

  this->m_board = new char*[this->m_height + 2];

  this->m_board[0] = new char[this->m_width + 2];
  memset(this->m_board[0], tile_impassable, this->m_width + 2);
  this->m_board[this->m_height + 1] = this->m_board[0];

  for (int row = 0; row < this->m_height; ++row)
    {
      char* r = new char[this->m_width + 2];
      r[0] = tile_impassable;

      // Read the newline
      sock.read(r + 1, this->m_width + 1);
      if (r[this->m_width + 1] != '\n')
	{
	  std::clog << "bad board length: " << r + 1 << std::endl;
	  exit(1);
	}
      r[this->m_width + 1] = tile_impassable;

      for (int col = 0; col < this->m_width; ++col)
	{
	  switch (r[col + 1])
	    {
	    case tile_plain:
	    case tile_lethal:
	    case tile_impassable:
	      break;

	    case tile_home:
	      this->m_home_tiles.push_back(Home_tile(col + 1, row + 1));
	      break;

	    default:
	      std::clog << "bad tile: " << r + 1 << std::endl;
	      exit(1);
	    }
	}

      this->m_board[row + 1] = r;
    }

  this->m_distances = new Distances*[this->m_height + 2];
  for (int row = 0; row < this->m_height + 2; ++row)
    { this->m_distances[row] = new Distances[this->m_width + 2]; }

  for (Home_tile_list::const_iterator p = this->m_home_tiles.begin();
       p != this->m_home_tiles.end();
       ++p)
    { this->find_distances(p->x(), p->y()); }
}

std::ostream&
operator<<(std::ostream& os, const Board& b)
{
  for (int r = 0; r < b.m_height + 2; ++r)
    {
      std::string s(b.m_board[r], b.m_width + 2);
      os << s << std::endl;
    }
  return os;
}

void
Board::check_neighbor(
  int x,
  int y,
  int xto,
  int yto,
  Point_list& pl) const
{
  if (this->m_distances[y][x].known(xto, yto))
    { return; }
  Tile t = this->at(x, y);
  if (t != tile_impassable && t != tile_lethal)
    pl.push_back(Xy(x, y));
}

void
Board::find_neighbors(
  int x,
  int y,
  int xto,
  int yto,
  Point_list& pl) const
{
  this->check_neighbor(x - 1, y, xto, yto, pl);
  this->check_neighbor(x + 1, y, xto, yto, pl);
  this->check_neighbor(x, y - 1, xto, yto, pl);
  this->check_neighbor(x, y + 1, xto, yto, pl);
}

void
Board::find_distances(
  int xto,
  int yto)
{
  if (this->m_distances[yto][xto].to(xto, yto) == 0)
    { return; }

  this->m_distances[yto][xto].add(xto, yto, 0);

  Point_list pl1;
  Point_list pl2;

  pl1.push_back(Xy(xto, yto));

  Point_list* ppl = &pl1;
  Point_list* pnpl = &pl2;

  int distance = 1;

  while (! ppl->empty())
    {
      while (! ppl->empty())
	{
	  Xy p = ppl->front();
	  ppl->pop_front();

	  Point_list neighbors;
	  this->find_neighbors(p.first, p.second, xto, yto, neighbors);
	  for (Point_list::const_iterator pn = neighbors.begin();
	       pn != neighbors.end();
	       ++pn)
	    {
	      this->m_distances[pn->second][pn->first].add(xto, yto, distance);
	    }
	  pnpl->splice(pnpl->end(), neighbors);
	}

      Point_list* pt = ppl;
      ppl = pnpl;
      pnpl = pt;

      pnpl->clear();
      ++distance;
    }

  if (debug_flag)
    {
      std::cout << "Distances to " << xto << ' ' << yto << std::endl;
      for (int r = 0; r < this->m_height + 2; ++r)
	{
	  for (int c = 0; c < this->m_width + 2; ++c)
	    {
	      std::cout << std::setw(3)
			<< this->m_distances[r][c].to(xto, yto);
	    }
	  std::cout << std::endl;
	}
    }
}

int
Board::distance(
  int xfrom,
  int yfrom,
  int xto,
  int yto)
{
  if (this->m_distances[yto][xto].to(xto, yto) != 0)
    { this->find_distances(xto, yto); }
  return this->m_distances[yfrom][xfrom].to(xto, yto);
}

void
Board::direction(
  int xfrom,
  int yfrom,
  int xto,
  int yto,
  bool dirs[direction_max])
{
  if (this->m_distances[yto][xto].to(xto, yto) != 0)
    { this->find_distances(xto, yto); }

  for (Direction d = direction_north; d < direction_max; ++d)
    { dirs[d] = false; }

  int distance = this->m_distances[yfrom][xfrom].to(xto, yto);
  if (distance == -1)
    {
      std::clog << "can't get to " << xto << ' ' << yto << " from "
		<< xfrom << ' ' << yfrom;
      return;
    }

  if (distance == 0)
    { return; }

  int d = this->m_distances[yfrom][xfrom - 1].to(xto, yto);
  if (d >= 0 && d < distance)
    { dirs[direction_west] = true; }
  d = this->m_distances[yfrom][xfrom + 1].to(xto, yto);
  if (d >= 0 && d < distance)
    { dirs[direction_east] = true; }
  d = this->m_distances[yfrom - 1][xfrom].to(xto, yto);
  if (d >= 0 && d < distance)
    { dirs[direction_south] = true; }
  d = this->m_distances[yfrom + 1][xfrom].to(xto, yto);
  if (d >= 0 && d < distance)
    { dirs[direction_north] = true; }
}

int
Board::lethal_direction(
  int x,
  int y,
  bool dirs[direction_max]) const
{
  int ret = 0;
  for (Direction d = direction_north; d < direction_max; ++d)
    { dirs[d] = false; }
  if (this->at(x - 1, y) == tile_lethal)
    {
      dirs[direction_west] = true;
      ++ret;
    }
  if (this->at(x + 1, y) == tile_lethal)
    {
      dirs[direction_east] = true;
      ++ret;
    }
  if (this->at(x, y - 1) == tile_lethal)
    {
      dirs[direction_south] = true;
      ++ret;
    }
  if (this->at(x, y + 1) == tile_lethal)
    {
      dirs[direction_north] = true;
      ++ret;
    }
  return ret;
}

int
Board::valid_direction(
  int x,
  int y,
  bool dirs[direction_max]) const
{
  int ret = 0;
  for (Direction d = direction_north; d < direction_max; ++d)
    { dirs[d] = false; }
  Tile t = this->at(x - 1, y);
  if (t != tile_lethal && t != tile_impassable)
    {
      dirs[direction_west] = true;
      ++ret;
    }
  t = this->at(x + 1, y);
  if (t != tile_lethal && t != tile_impassable)
    {
      dirs[direction_east] = true;
      ++ret;
    }
  t = this->at(x, y - 1);
  if (t != tile_lethal && t != tile_impassable)
    {
      dirs[direction_south] = true;
      ++ret;
    }
  t = this->at(x, y + 1);
  if (t != tile_lethal && t != tile_impassable)
    {
      dirs[direction_north] = true;
      ++ret;
    }
  return ret;
}

void
Board::saw_tile(
  int x,
  int y)
{
  for (Home_tile_list::iterator p = this->m_home_tiles.begin();
       p != this->m_home_tiles.end();
       ++p)
    {
      if (p->x() == x && p->y() == y)
	{ p->known(); }
    }
}

void
Board::check_neighbor(
  int x,
  int y,
  Point_list& all,
  Point_list& pl) const
{
  if (std::find(all.begin(), all.end(), Xy(x, y)) != all.end())
    { return; }
  Tile t = this->at(x, y);
  if (t != tile_impassable && t != tile_lethal)
    {
      pl.push_back(Xy(x, y));
      all.push_back(Xy(x, y));
    }
}

void
Board::find_neighbors(
  int x,
  int y,
  Point_list& all,
  Point_list& pl) const
{
  this->check_neighbor(x - 1, y, all, pl);
  this->check_neighbor(x + 1, y, all, pl);
  this->check_neighbor(x, y - 1, all, pl);
  this->check_neighbor(x, y + 1, all, pl);
}

bool
Board::find(
  int x, 
  int y,
  Find& fn,
  int max,
  int& xret,
  int& yret,
  int& distanceret)
{
  int distance = 0;

  Point_list all;

  Point_list pl1;
  Point_list pl2;

  all.push_back(Xy(x, y));
  pl1.push_back(Xy(x, y));

  Point_list* ppl = &pl1;
  Point_list* pnpl = &pl2;

  while ((max == 0 || distance <= max) && ! ppl->empty())
    {
      while (! ppl->empty())
	{
	  Xy p = ppl->front();
	  ppl->pop_front();

	  if (fn(*this, p.first, p.second, distance))
	    {
	      xret = p.first;
	      yret = p.second;
	      distanceret = distance;
	      return true;
	    }

	  find_neighbors(p.first, p.second, all, *pnpl);
	}

      Point_list* pt = ppl;
      ppl = pnpl;
      pnpl = pt;

      pnpl->clear();
      ++distance;
    }

  return false;
}
