// board.h  -*- C++ -*-
// Ian Lance Taylor for ICFP 2002

// A class to represent the game board.

#ifndef BOARD_H
#define BOARD_H

#include <map>
#include <iostream>

#include "socket.h"
#include "package.h"

// Each tile on the board has a list of distances to other tiles of
// interest.

enum Direction
  {
    direction_north,
    direction_south,
    direction_east,
    direction_west,
    direction_max
  };

// Adjust x and y for the given direction.
extern void in_direction(Direction, int& x, int& y);

// Return the direction for a character.
extern Direction direction_for_char(char);

// Return the character for a direction.
extern char char_for_direction(Direction);

// Return the reverse of a direction.
extern Direction reverse_direction(Direction);

// Return the direction to the right.
extern Direction direction_right(Direction);

inline Direction operator++(Direction& d)
{ return d = static_cast<Direction>(static_cast<int>(d) + 1); }

extern std::ostream& operator<<(std::ostream&, Direction);

class Distances
{
public:
  Distances()
    : m_distances()
  { }

  void add(int x, int y, int distance)
  { this->m_distances[std::make_pair(x, y)] = distance; }

  bool known(int x, int y) const;

  int to(int x, int y) const;

private:
  typedef std::map<Xy, int> Distance_map;

  Distance_map m_distances;
};

// For home tiles, we record whether we have been to them.

class Home_tile
{
public:
  Home_tile(int x, int y)
    : m_x(x), m_y(y), m_known(false)
  { }

  int x() const
  { return this->m_x; }
  int y() const
  { return this->m_y; }
  bool is_known() const
  { return this->m_known; }

  void known()
  { this->m_known = true; }

private:
  int m_x;
  int m_y;
  bool m_known;
};

typedef std::list<Home_tile> Home_tile_list;

class Board
{
public:
  Board()
    : m_width(0), m_height(0), m_board(0), m_distances(), m_home_tiles()
  { }

  enum Tile
    {
      tile_plain = '.',
      tile_lethal = '~',
      tile_impassable = '#',
      tile_home = '@'
    };

  void read(Socket&);

  Tile at(int x, int y) const
  { return static_cast<Tile>(this->m_board[y][x]); }

  void find_distances(int x, int y);

  int distance(int xfrom, int yfrom, int xto, int yto);
  void direction(int xfrom, int yfrom, int xto, int yto,
		 bool dirs[direction_max]);

  // Set dirs according to which directions we can find a lethal tile.
  // Return the number found.
  int lethal_direction(int x, int y, bool dirs[direction_max]) const;

  // Set dirs to directions in which we are permitted to move.  Return
  // the number found.
  int valid_direction(int x, int y, bool dirs[direction_max]) const;

  void saw_tile(int x, int y);

  friend std::ostream& operator<<(std::ostream&, const Board&);

  Home_tile_list::const_iterator home_tile_begin()
  { return this->m_home_tiles.begin(); }
  Home_tile_list::const_iterator home_tile_end()
  { return this->m_home_tiles.end(); }

  // Find the nearest point which satisfies a function, up to a
  // maximum distance (0 for no maximum).  Return whether any such
  // point was found.  Set the point and the distance in the reference
  // variables.
  class Find
  {
  public:
    virtual bool operator()(Board&, int x, int y, int distance) = 0;
  };
  bool find(int x, int y, Find&, int max, int& xret, int& yret, int& distance);

private:
  typedef std::list<Xy> Point_list;

  void check_neighbor(int x, int y, int xto, int yto, Point_list&) const;
  void find_neighbors(int x, int y, int xto, int yto, Point_list&) const;
  void check_neighbor(int x, int y, Point_list&, Point_list&) const;
  void find_neighbors(int x, int y, Point_list&, Point_list&) const;

  int m_width;
  int m_height;
  char** m_board;
  Distances** m_distances;
  Home_tile_list m_home_tiles;
};

#endif // ! defined(BOARD_H)
