// main.cc
// Ian Lance Taylor for ICFP 2002

// Main program.

#include <iostream>
#include <cstdlib>
#include <ctime>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>

#include "debug.h"
#include "socket.h"
#include "board.h"
#include "player.h"
#include "robots.h"
#include "package.h"
#include "state.h"
#include "strategy.h"

bool debug_flag = false;

int
main(
  int argc,
  char**argv)
{
  if (argc != 3 && argc != 4)
    {
      std::clog << "Usage: runme host port [debug]" << std::endl;
      exit(1);
    }

  if (argc == 4)
    { debug_flag = true; }

  srandom(time(0));

  // The strategy list is built at global constructor time, but it is
  // unsorted.
  strategies.sort();

  char const* host = argv[1];
  char const* port = argv[2];

  addrinfo hints;
  memset(&hints, 0, sizeof hints);
  hints.ai_socktype = SOCK_STREAM;
  addrinfo *addr;
  int ret = getaddrinfo(host, port, &hints, &addr);
  if (ret != 0)
    {
      std::clog << "getaddrinfo failed: " << gai_strerror(ret) << std::endl;
      exit(1);
    }

  int o = -1;
  int ierr = 0;
  for (addrinfo* p = addr; p != 0; p = p->ai_next)
    {
      o = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
      if (o >= 0)
	{
	  if (connect(o, p->ai_addr, p->ai_addrlen) >= 0)
	    { break; }
	  ierr = errno;
	  close(o);
	  o = -1;
	}
    }

  if (o < 0)
    {
      if (ierr == 0)
	std::clog << "getaddrinfo returned no useful address" << std::endl;
      else
	std::clog << "connect: " << strerror(ierr) << std::endl;
      exit(1);
    }

  Socket s(o);
  s.write("Player\n");

  Board board;
  board.read(s);
  if (debug_flag)
    { std::cout << board; }

  Player me;
  me.read(s);
  if (debug_flag)
    { std::cout << me; }

  Packages packs;

  Robots robots;
  robots.read(me, s, packs);
  if (debug_flag)
    { std::cout << robots; }

  int xe;
  int ye;
  me.location(robots, xe, ye);
  Package_list pe;

  while (true)
    {
      State state(board, me, robots, packs);

      bool bumped = (xe != state.xme()
		     || ye != state.yme()
		     || pe != me.packages(robots));
      me.bumped(bumped);

      Package_list pl = packs.read(me, robots, s, state.xme(), state.yme());

      board.saw_tile(state.xme(), state.yme());

      Actions actions;
      for (Strategy_list::const_iterator ps = strategies.begin();
	   ps != strategies.end();
	   ++ps)
	{ (**ps)(state, pl, actions); }

      int bid;
      std::string cmd = actions.command(state, bid, xe, ye, pe);
      if (debug_flag)
	{ std::cout << cmd; }
      s.write(cmd);

      me.pay(bid);

      robots.read(me, s, packs);
      if (debug_flag)
	{ std::cout << robots; }
    }
}
