// package.cc
// Ian Lance Taylor for ICFP 2002

// Keep information about packages.

#include <cstdlib>
#include <iostream>
#include <sstream>

#include "debug.h"
#include "socket.h"
#include "player.h"
#include "board.h"
#include "package.h"

bool
Package::dest(
  int& x,
  int& y) const
{
  if (this->m_dest_x == -1)
    { return false; }
  x = this->m_dest_x;
  y = this->m_dest_y;
  return true;
}

void
Package::at(
  int x,
  int y)
{
  this->m_x = x;
  this->m_y = y;
  this->m_robot = -1;
}

void
Package::with(
  int robot)
{
  this->m_x = -1;
  this->m_y = -1;
  this->m_robot = robot;
}

void
Package::info(
  int x,
  int y,
  int dest_x,
  int dest_y,
  int weight)
{
  this->m_x = x;
  this->m_y = y;
  this->m_robot = -1;
  this->m_dest_x = dest_x;
  this->m_dest_y = dest_y;
  this->m_weight = weight;
}

static void
package_error(
  const std::string& s)
{
  std::clog << "package list syntax error: " << s << std::endl;
  exit(1);
}

Package_list
Packages::read(
  const Player& me,
  const Robots& robots,
  Socket& sock,
  int x,
  int y)
{
  // Forget any knowledge of packages at this location.  Recreate the
  // knowledge based on what the response tell us.  Our knowledge can
  // be inaccurate if another robot dropped a package at its
  // destination when we didn't know the package destination.
  Package_locs::iterator ppl = this->m_package_locs.find(Xy(x, y));
  if (ppl != this->m_package_locs.end())
    { this->m_package_locs.erase(ppl); }

  Package_list ret;
  std::string s = sock.getline();

  if (debug_flag)
    { std::cout << s << std::endl; }

  const char* p = s.c_str();
  while (*p != '\0')
    {
      char* end;
      int pid = strtol(p, &end, 10);
      p = end;
      if (*p != ' ')
	{
	  // When the robot has delivered all packages, the test
	  // server sends "Robot N died."
	  std::ostringstream oss;
	  oss << "Robot " << me.id() << " died.";
	  if (s == oss.str())
	    {
	      std::cout << "Score " << me.score(robots) << std::endl;
	      exit(0);
	    }

	  package_error(s);
	}

      int dest_x = strtol(p, &end, 10);
      p = end;
      if (*p != ' ')
	{ package_error(s); }

      int dest_y = strtol(p, &end, 10);
      p = end;
      if (*p != ' ')
	{ package_error(s); }

      int weight = strtol(p, &end, 10);
      p = end;

      if (pid < 0)
	{ package_error(s); }

      if (static_cast<unsigned int>(pid) >= this->m_packages.size())
	{ this->m_packages.resize(pid + 1); }

      this->m_packages[pid].info(x, y, dest_x, dest_y, weight);
      this->m_package_locs[Xy(x, y)].insert(pid);
      ret.insert(pid);

      if (*p == ' ')
	{ ++p; }
    }

  return ret;
}

int
Packages::weight(
  int pid) const
{
  assert(pid >= 0);
  if (static_cast<unsigned int>(pid) >= this->m_packages.size())
    { return -1; }
  return this->m_packages[pid].weight();
}

bool
Packages::dest(
  int pid,
  int& x,
  int& y) const
{
  assert(pid >= 0);
  if (static_cast<unsigned int>(pid) >= this->m_packages.size())
    { return false; }
  return this->m_packages[pid].dest(x, y);
}

Xy
Packages::dest(
  int pid) const
{
  assert(pid >= 0);
  if (static_cast<unsigned int>(pid) >= this->m_packages.size())
    { return Xy(-1, -1); }
  return this->m_packages[pid].dest();
}

void
Packages::at(
  int pid,
  int x,
  int y)
{
  assert(pid >= 0);
  if (static_cast<unsigned int>(pid) >= this->m_packages.size())
    { this->m_packages.resize(pid + 1); }

  int xdest;
  int ydest;
  if (this->m_packages[pid].dest(xdest, ydest)
      && xdest == x
      && ydest == y)
    { this->m_packages[pid] = Package(); }
  else
    {
      this->m_packages[pid].at(x, y);
      this->m_package_locs[Xy(x, y)].insert(pid);
    }
}

void
Packages::with(
  int pid,
  int rid)
{
  assert(pid >= 0);
  if (static_cast<unsigned int>(pid) >= this->m_packages.size())
    { this->m_packages.resize(pid + 1); }

  if (this->m_packages[pid].xy_is_known())
    { this->m_package_locs[this->m_packages[pid].xy()].erase(pid); }

  this->m_packages[pid].with(rid);
}

size_t
Packages::count(
  int x,
  int y) const
{
  Package_locs::const_iterator ppl = this->m_package_locs.find(Xy(x, y));
  if (ppl == this->m_package_locs.end())
    { return 0; }
  return ppl->second.size();
}

bool
Packages::nearest_package(
  Board& board,
  int x,
  int y,
  int& xnearest,
  int& ynearest) const
{
  int distance = -1;
  int xn = 0;
  int yn = 0;
  for (Package_locs::const_iterator ppl = this->m_package_locs.begin();
       ppl != this->m_package_locs.end();
       ++ppl)
    {
      int d = board.distance(x, y, ppl->first.first, ppl->first.second);
      if (distance == -1 || d < distance)
	{
	  distance = d;
	  xn = ppl->first.first;
	  yn = ppl->first.second;
	}
    }

  if (distance == -1)
    { return false; }

  xnearest = xn;
  ynearest = yn;
  return true;
}

std::ostream&
operator<<(
  std::ostream& os,
  const Package_list& pl)
{
  bool first = true;
  for (Package_list::const_iterator p = pl.begin();
       p != pl.end();
       ++p)
    {
      if (first)
	{ first = false; }
      else
	{ os << ' '; }
      os << *p;
    }
  return os;
}
