// robots.cc
// Ian Lance Taylor for ICFP 2002

// Implementation of list of robots.

#include <cassert>
#include <cstdlib>
#include <iostream>
#include <sstream>

#include "debug.h"
#include "socket.h"
#include "package.h"
#include "board.h"
#include "player.h"
#include "robots.h"

int Robot::turn = 0;

void
Robot::set(
  int x,
  int y)
{
  this->m_x = x;
  this->m_y = y;
}

void
Robot::move(
  Direction d)
{
  in_direction(d, this->m_x, this->m_y);
}

void
Robot::pickup(
  int pid)
{
  this->m_packages.insert(pid);
}

void
Robot::drop(
  int pid)
{
  this->m_packages.erase(pid);
}

void
Robot::location(
  int& x,
  int& y) const
{
  x = this->m_x;
  y = this->m_y;
}

int
Robot::weight(
  const Packages& packs) const
{
  int ret = 0;
  for (Package_list::const_iterator p = this->m_packages.begin();
       p != this->m_packages.end();
       ++p)
    {
      int w = packs.weight(*p);
      if (w > 0)
	{ ret += w; }
    }
  return ret;
}

std::ostream&
operator<<(
  std::ostream& os,
  const Robot& r)
{
  os << r.m_x << ' ' << r.m_y;
  for (Package_list::const_iterator p = r.m_packages.begin();
       p != r.m_packages.end();
       ++p)
    { os << ' ' << *p; }
  os << std::endl;
  return os;
}

static void
response_error(
  const Player& me,
  const Robots& robots,
  const std::string& s)
{
  // When the robot runs out of money, the test server sends "Robot N
  // died."
  std::ostringstream oss;
  oss << "Robot " << me.id() << " died.";
  if (s == oss.str())
    {
      std::cout << "Score " << me.score(robots) << std::endl;
      exit(0);
    }

  std::clog << "response syntax error: " << s << std::endl;
  exit(1);
}

void
Robots::read(
  const Player& me,
  Socket& sock,
  Packages& packs)
{
  Robot::change_turn();

  std::string s = sock.getline();
  if (debug_flag)
    { std::cout << s << std::endl; }
  const char* p = s.c_str();
  while (*p != '\0')
    {
      if (*p != '#')
	{ response_error(me, *this, s); }
      ++p;

      char* end;
      int id = strtol(p, &end, 10);
      p = end;
      if (id < 0 || (*p != ' ' && *p != '\0'))
	{ response_error(me, *this, s); }
      if (*p == ' ')
	{ ++p; }

      if (*p != 'X')
	{
	  if (static_cast<unsigned int>(id) >= this->m_robots.size()
	      || ! this->m_robots[id].was_valid())
	    {
	      std::clog << "reference to unknown robot " << id << std::endl;
	    }

	  this->m_robots[id].mark_valid();
	}

      while (*p != '#' && *p != '\0')
	{
	  switch (*p)
	    {
	    case 'N':
	    case 'S':
	    case 'E':
	    case 'W':
	      this->m_robots[id].move(direction_for_char(*p));
	      ++p;
	      break;

	    case 'P':
	    case 'D':
	      if (p[1] != ' ')
		{ response_error(me, *this, s); }
	      {
		int pid = strtol(p + 2, &end, 10);
		if (*p == 'P')
		  {
		    this->m_robots[id].pickup(pid);
		    packs.with(pid, id);
		  }
		else
		  {
		    this->m_robots[id].drop(pid);

		    int x;
		    int y;
		    if (packs.dest(pid, x, y)
			&& x == this->m_robots[id].x()
			&& y == this->m_robots[id].y())
		      {
			int w = packs.weight(pid);
			if (w >= 0)
			  { this->m_robots[id].scored(w); }
		      }

		    packs.at(pid, this->m_robots[id].x(),
			     this->m_robots[id].y());
		  }
		p = end;
	      }
	      break;

	    case 'X':
	      if (p[1] != ' ')
		{ response_error(me, *this, s); }
	      {
		int x = strtol(p + 2, &end, 10);
		p = end;
		if (*p != ' ' || p[1] != 'Y' || p[2] != ' ')
		  { response_error(me, *this, s); }
		int y = strtol(p + 3, &end, 10);
		if (static_cast<unsigned int>(id) >= this->m_robots.size())
		  { this->m_robots.resize(id + 1); }
		this->m_robots[id].set(x, y);
		this->m_robots[id].mark_valid();
		p = end;
	      }
	      break;
	    }

	  if (*p == ' ')
	    { ++p; }
	}
    }
}

void
Robots::location(
  int id,
  int& x,
  int& y) const
{
  assert(id >= 0 && static_cast<unsigned int>(id) < this->m_robots.size());
  if (! this->m_robots[id].is_valid())
    {
      std::clog << "request for location of invalid robot: " << id
		<< std::endl;
      return;
    }
  this->m_robots[id].location(x, y);
}

Xy
Robots::xy(
  int id) const
{
  assert(id >= 0 && static_cast<unsigned int>(id) < this->m_robots.size());
  if (! this->m_robots[id].is_valid())
    {
      std::clog << "request for location of invalid robot: " << id
		<< std::endl;
      return Xy(-1, -1);
    }
  return this->m_robots[id].xy();
}

int
Robots::score(
  int id) const
{
  assert(id >= 0 && static_cast<unsigned int>(id) < this->m_robots.size());
  if (! this->m_robots[id].is_valid())
    {
      std::clog << "request for score of invalid robot: " << id
		<< std::endl;
      return 0;
    }
  return this->m_robots[id].score();
}

Package_list
Robots::packages(
  int id) const
{
  assert(id >= 0 && static_cast<unsigned int>(id) < this->m_robots.size());
  if (! this->m_robots[id].is_valid())
    {
      std::clog << "request for packages of invalid robot: " << id
		<< std::endl;
      return Package_list();
    }
  return this->m_robots[id].packages();
}

int
Robots::weight(
  int id,
  const Packages& packs) const
{
  assert(id >= 0 && static_cast<unsigned int>(id) < this->m_robots.size());
  if (! this->m_robots[id].is_valid())
    {
      std::clog << "request for weight of invalid robot: " << id
		<< std::endl;
      return 0;
    }
  return this->m_robots[id].weight(packs);
}

int
Robots::at(
  int x,
  int y) const
{
  for (std::vector<Robot>::const_iterator p = this->m_robots.begin();
       p != this->m_robots.end();
       ++p)
    {
      if (p->is_valid() && p->x() == x && p->y() == y)
	{ return p - this->m_robots.begin(); }
    }
  return -1;
}

std::ostream&
operator<<(
  std::ostream& os,
  const Robots& rs)
{
  for (std::vector<Robot>::const_iterator p = rs.m_robots.begin();
       p != rs.m_robots.end();
       ++p)
    {
      if (p->is_valid())
	{ os << p - rs.m_robots.begin() << ' ' << *p; }
    }
  return os;
}
