// socket.cc
// Ian Lance Taylor for ICFP 2002

// Socket implementation.

#include <cstring>
#include <cerrno>
#include <cstdlib>
#include <iostream>
#include <string>
#include <list>
#include <unistd.h>

#include "debug.h"
#include "socket.h"

const size_t Buffers::page_size = ::sysconf(_SC_PAGESIZE);

void
Buffers::clear()
{
  for (Buffer_list::iterator p = this->m_buffers.begin();
       p != this->m_buffers.end();
       ++p)
    { free(p->alc); }
  this->m_buffers.clear();
}

size_t
Buffers::size() const
{
  size_t ret = 0;
  for (Buffer_list::const_iterator p = this->m_buffers.begin();
       p != this->m_buffers.end();
       ++p)
    { ret += p->len; }
  return ret;
}

void
Buffers::add_buff(
  char* buff,
  size_t len)
{
  if (len == 0)
    { return; }
  Buffer b;
  b.alc = buff;
  b.data = b.alc;
  b.len = len;
  this->m_buffers.push_back(b);
}

void
Buffers::add(
  const void* void_buff,
  size_t len)
{
  const char* buff = static_cast<const char*>(void_buff);
  while (len > 0)
    {
      char* alc = static_cast<char*>(valloc(Buffers::page_size));
      size_t c = len;
      if (c > Buffers::page_size)
	{ c = Buffers::page_size; }
      memcpy(alc, buff, c);
      this->add_buff(alc, c);

      buff += c;
      len -= c;
    }
}

void
Buffers::read(
  int o)
{
  char* alc = static_cast<char*>(valloc(Buffers::page_size));

  ssize_t got = ::read(o, alc, Buffers::page_size);
  if (got < 0)
    {
      std::clog << "read: " << strerror(errno) << std::endl;
      exit(1);
    }
  if (got == 0)
    {
      if (debug_flag)
	{ std::clog << "read EOF" << std::endl; }
      exit(0);
    }

  this->add_buff(alc, got);
}

size_t
Buffers::consume(
  size_t len)
{
  size_t ret = 0;
  while (len > 0)
    {
      if (this->m_buffers.empty())
	{ return ret; }

      if (len < this->m_buffers.front().len)
	{
	  this->m_buffers.front().len -= len;
	  this->m_buffers.front().data += len;
	  ret += len;
	  return ret;
	}

      len -= this->m_buffers.front().len;
      ret += this->m_buffers.front().len;

      free(this->m_buffers.front().alc);
      this->m_buffers.pop_front();
    }

  return ret;
}

bool
Buffers::find(
  unsigned char c,
  size_t& loc) const
{
  loc = 0;
  for (Buffer_list::const_iterator p = this->m_buffers.begin();
       p != this->m_buffers.end();
       ++p)
    {
      const char* pc = static_cast<char*>(memchr(p->data, c, p->len));
      if (pc != 0)
	{
	  loc += pc - p->data + 1;
	  return true;
	}
      loc += p->len;
    }
  return false;
}

size_t
Buffers::move(
  void* void_buff,
  size_t len)
{
  char* buff = static_cast<char*>(void_buff);
  size_t ret = 0;
  while (len > 0)
    {
      if (this->m_buffers.empty())
	{ return ret; }

      size_t front_len = this->m_buffers.front().len;
      if (front_len > len)
	{
	  memcpy(buff, this->m_buffers.front().data, len);
	  this->m_buffers.front().len -= len;
	  this->m_buffers.front().data += len;
	  ret += len;
	  return ret;
	}

      memcpy(buff, this->m_buffers.front().data, front_len);

      len -= front_len;
      buff += front_len;
      ret += front_len;

      free(this->m_buffers.front().alc);
      this->m_buffers.pop_front();
    }

  return ret;
}

size_t
Buffers::move_string(
  std::string& s,
  size_t len)
{
  size_t ret = 0;
  while (len > 0)
    {
      if (this->m_buffers.empty())
	{ return ret; }

      size_t front_len = this->m_buffers.front().len;
      if (front_len > len)
	{
	  s.append(this->m_buffers.front().data, len);
	  this->m_buffers.front().len -= len;
	  this->m_buffers.front().data += len;
	  ret += len;
	  return ret;
	}

      s.append(this->m_buffers.front().data, front_len);

      len -= front_len;
      ret += front_len;

      free(this->m_buffers.front().alc);
      this->m_buffers.pop_front();
    }

  return ret;
}

void
Socket::write(
  const void* buff,
  size_t len)
{
  while (len > 0)
    {
      ssize_t wrote = ::write(this->m_o, buff, len);
      if (wrote < 0)
	{
	  std::clog << "write failed: " << strerror(errno) << std::endl;
	  exit(1);
	}
      if (wrote == 0)
	{
	  std::clog << "write EOF" << std::endl;
	  exit(1);
	}

      len -= wrote;
    }
}

void
Socket::write(
  const char* str)
{
  this->write(str, strlen(str));
}

void
Socket::write(
  const std::string& s)
{
  this->write(s.data(), s.length());
}

void
Socket::writeln(
  const char* str)
{
  std::string s = std::string(str) + '\n';
  this->write(s);
}

void
Socket::writeln(
  const std::string& s)
{
  std::string s1 = s + '\n';
  this->write(s);
}

void
Socket::read(
 void* buff,
 size_t len)
{
  while (this->m_buffers.size() < len)
    { this->m_buffers.read(this->m_o); }
  this->m_buffers.move(buff, len);
}

std::string
Socket::getline()
{
  size_t loc;
  while (! this->m_buffers.find('\n', loc))
    { this->m_buffers.read(this->m_o); }

  std::string s;
  s.reserve(loc - 1);
  this->m_buffers.move_string(s, loc - 1);
  this->m_buffers.consume(1);

  return s;
}
