// strategy.cc
// Ian Lance Taylor for ICFP 2002

// Implement strategy and action functions.

#include <cassert>
#include <cstdlib>
#include <string>
#include <map>
#include <iostream>
#include <sstream>

#include "debug.h"
#include "board.h"
#include "player.h"
#include "robots.h"
#include "strategy.h"

Strategy_list strategies;

void
Actions::adjust_move(
  Direction d,
  int adjust,
  Time time)
{
  if (debug_flag)
    { std::cout << "Adjust direction " << d << " + " << adjust << std::endl; }
  this->m_move[d][time] += adjust;
}

void
Actions::adjust_pickup(
  const Package_list& pl,
  int adjust,
  Time time)
{
  if (debug_flag)
    { std::cout << "Adjust pickup " << pl << " + " << adjust << std::endl; }
  Package_score::iterator p = this->m_pickup.find(pl);
  if (p != this->m_pickup.end())
    { p->second.scores[time] += adjust; }
  else
    {
      int* pt = this->m_pickup[pl].scores;
      memset(pt, 0, sizeof(int[time_max]));
      pt[time] = adjust;
    }
}

void
Actions::adjust_drop(
  const Package_list& pl,
  int adjust,
  Time time)
{
  if (debug_flag)
    { std::cout << "Adjust drop " << pl << " + " << adjust << std::endl; }
  Package_score::iterator p = this->m_drop.find(pl);
  if (p != this->m_drop.end())
    { p->second.scores[time] += adjust; }
  else
    {
      int* pt = this->m_drop[pl].scores;
      memset(pt, 0, sizeof(int[time_max]));
      pt[time] = adjust;
    }
}

void
Actions::adjust_move_to(
  Board& board,
  const Player& me,
  const Robots& robots,
  int x,
  int y,
  int adjust,
  Time time)
{
  if (debug_flag)
    {
      std::cout << "Adjust move to " << x << ' ' << y << " + " << adjust
		<< std::endl;
    }

  int xhere;
  int yhere;
  me.location(robots, xhere, yhere);

  bool dirs[direction_max];
  board.direction(xhere, yhere, x, y, dirs);

  for (Direction d = direction_north; d < direction_max; ++d)
    {
      if (dirs[d])
	{ this->m_move[d][time] += adjust; }
    }
}

void
Actions::get_score(
  const int* pt,
  int& score,
  Time& time) const
{
  if (pt[time_first] > 0 && pt[time_first] > pt[time_last])
    {
      score = pt[time_first] + pt[time_any];
      time = time_first;
    }
  else if (pt[time_last] > 0)
    {
      score = pt[time_last] + pt[time_any];
      time = time_last;
    }
  else
    {
      score = pt[time_any];
      time = time_any;
    }
}

Direction
Actions::best_move() const
{
  int max = 0;
  Direction max_move = direction_max;
  for (Direction d = direction_north; d < direction_max; ++d)
    {
      int score;
      Time time;
      this->get_score(this->m_move[d], score, time);
      if (score > max)
	{
	  max = score;
	  max_move = d;
	}
    }
  return max_move;
}

std::string
Actions::command(State& state, int& bid, int& xe, int& ye, Package_list& pe)
{
  int max = 0;
  Time max_time = time_any;
  Direction max_move = direction_max;
  const Package_list* pickup = 0;
  const Package_list* drop = 0;
  int score;
  Time time;

  for (Direction d = direction_north; d < direction_max; ++d)
    {
      this->get_score(this->m_move[d], score, time);
      if (score > max)
	{
	  max = score;
	  max_time = time;
	  max_move = d;
	}
    }

  for (Package_score::const_iterator p = this->m_pickup.begin();
       p != this->m_pickup.end();
       ++p)
    {
      this->get_score(p->second.scores, score, time);
      if (score > max)
	{
	  max = score;
	  max_time = time;
	  max_move = direction_max;
	  pickup = &p->first;
	}
    }

  for (Package_score::const_iterator p = this->m_drop.begin();
       p != this->m_drop.end();
       ++p)
    {
      this->get_score(p->second.scores, score, time);
      if (score > max)
	{
	  max = score;
	  max_time = time;
	  max_move = direction_max;
	  pickup = 0;
	  drop = &p->first;
	}
    }

  xe = state.xme();
  ye = state.yme();
  pe = state.me().packages(state.robots());

  std::ostringstream oss;
  switch (max_time)
    {
      // FIXME: There should be more control over these scores, and we
      // should check the amount of money we have left.
    case time_first:
      oss << "10 ";
      bid = 10;
      break;

    case time_last:
      oss << "-10 ";
      bid = 10;
      break;

    case time_any:
      oss << "1 ";
      bid = 1;
      break;

    default:
      assert(0);
    }

  if (max_move != direction_max)
    {
      oss << "Move " << max_move;
      in_direction(max_move, xe, ye);
    }
  else if (pickup != 0 || drop != 0)
    {
      const Package_list* ppl;
      if (pickup != 0)
	{
	  oss << "Pick";
	  ppl = pickup;
	  pe.insert(pickup->begin(), pickup->end());
	}
      else
	{
	  oss << "Drop";
	  ppl = drop;
	  for (Package_list::const_iterator p = drop->begin();
	       p != drop->end();
	       ++p)
	    { pe.erase(*p); }
	}

      for (Package_list::const_iterator p = ppl->begin();
	   p != ppl->end();
	   ++p)
	{ oss << ' ' << *p; }
    }
  else
    {
      // We can't think of anything to do.
      oss << "Pick";
    }

  oss << '\n';

  return oss.str();
}

// Basic strategies.

// Value of moving toward nearest home location which has packages.
const int value_home_nearest = 6;

// Value of moving toward nearest home location if none have packages.
const int value_empty_home_nearest = 2;

// Value of moving toward nearest location which has packages.
const int value_package_nearest = 4;

// Value of picking up packages;
const int value_pickup = 40;

// Value of moving toward package destination.
const int value_move_to_drop = 15;

// Value of dropping packages at their destination.
const int value_drop = 50;

// Value of moving toward robots.  This is negative so that it
// actually causes us to move away from robots.
const int value_meet_robot = -2;

// Value of staying dry--avoiding getting caught between another robot
// and water.
const int value_dry_robot = 3;

// Value of staying dry if we are in immediate danger.
const int value_dry_endangered_robot = 100;

// Value of avoiding a bump if we are getting bumped a lot.
const int value_avoid_bump = 25;

// Value of avoiding a bump if we are getting bumped all the time.
const int value_avoid_many_bumps = 45;

// Value of moving randomly if we have nothing else to do.
const int value_random = 3;

// Value of opportunistically killing another robot.
const int value_kill_robot = 30;

// Move toward the nearest home location which has, or may have,
// packages.

class Home_strategy : public Strategy
{
  void operator()(State&, Package_list&, Actions&) const;
};

void
Home_strategy::operator()(
  State& state,
  Package_list&,
  Actions& acts) const
{
  int xnearest = -1;
  int ynearest = -1;
  int nearest_distance = 0;

  int value = value_home_nearest;

  for (Home_tile_list::const_iterator p = state.board().home_tile_begin();
       p != state.board().home_tile_end();
       ++p)
    {
      if (! p->is_known()
	  || state.packs().count(p->x(), p->y()) > 0)
	{
	  int distance = state.board().distance(state.xme(), state.yme(),
						p->x(), p->y());
	  if (distance > 0
	      && (xnearest == -1 || distance < nearest_distance))
	    {
	      xnearest = p->x();
	      ynearest = p->y();
	      nearest_distance = distance;
	    }
	}
    }

  if (xnearest == -1)
    {
      value = value_empty_home_nearest;

      for (Home_tile_list::const_iterator p = state.board().home_tile_begin();
	   p != state.board().home_tile_end();
	   ++p)
	{
	  int distance = state.board().distance(state.xme(), state.yme(),
						p->x(), p->y());
	  if (distance > 0
	      && (xnearest == -1 || distance < nearest_distance))
	    {
	      xnearest = p->x();
	      ynearest = p->y();
	      nearest_distance = distance;
	    }
	}
    }

  if (xnearest != -1)
    {
      acts.adjust_move_to(state.board(), state.me(), state.robots(),
			  xnearest, ynearest, value);
    }
}

Home_strategy home_strategy;

Add_strategy add_home_strategy(home_strategy);

// Move toward the nearest location which has a package.

class Nearest_package_strategy : public Strategy
{
  void operator()(State&, Package_list&, Actions&) const;
};

void
Nearest_package_strategy::operator()(
  State& state,
  Package_list&,
  Actions& acts) const
{
  int xnearest;
  int ynearest;
  if (! state.packs().nearest_package(state.board(), state.xme(), state.yme(),
				      xnearest, ynearest))
    { return; }

  acts.adjust_move_to(state.board(), state.me(), state.robots(),
		      xnearest, ynearest, value_package_nearest);
}

Nearest_package_strategy nearest_package_strategy;

Add_strategy add_nearest_package_strategy(nearest_package_strategy);

// If there are any packages on the current square, pick up all that
// we can carry.

class
Pickup_strategy : public Strategy
{
  void operator()(State&, Package_list&, Actions&) const;
};

class Pack_id_weight
{
public:
  Pack_id_weight(int id, int weight)
    : m_id(id), m_weight(weight)
  { }

  int id() const
  { return this->m_id; }
  int weight() const
  { return this->m_weight; }

private:
  int m_id;
  int m_weight;
};

// Use a reverse ordering so that the heaviest objects come first.
inline bool
operator<(const Pack_id_weight& p1, const Pack_id_weight& p2)
{
  return p1.weight() > p2.weight();
}

void
Pickup_strategy::operator()(
  State& state,
  Package_list& pl,
  Actions& acts) const
{
  if (pl.empty())
    { return; }

  int capacity = (state.me().carry()
		  - state.me().weight(state.robots(), state.packs()));

  std::list<Pack_id_weight> piwl;
  for (Package_list::const_iterator p = pl.begin();
       p != pl.end();
       ++p)
    { piwl.push_back(Pack_id_weight(*p, state.packs().weight(*p))); }
  piwl.sort();

  Package_list have = state.me().packages(state.robots());

  Package_list pickup;

  for (int pass = 0; pass < 2; ++pass)
    {
      std::list<Pack_id_weight>::iterator p = piwl.begin();
      while (p != piwl.end())
	{
	  if (piwl.front().weight() > capacity)
	    {
	      ++p;
	      continue;
	    }

	  // In the first pass, only take packages with the same
	  // destination as some other package we already have.
	  if (pass == 0
	      && (! have.empty() || ! pickup.empty()))
	    {
	      Package_list::const_iterator ppl;
	      for (ppl = have.begin(); ppl != have.end(); ++ppl)
		{
		  if (state.packs().dest(*ppl)
		      == state.packs().dest(piwl.front().id()))
		    { break; }
		}
	      if (ppl == have.end())
		{
		  for (ppl = pickup.begin(); ppl != pickup.end(); ++ppl)
		    {
		      if (state.packs().dest(*ppl)
			  == state.packs().dest(piwl.front().id()))
			{ break; }
		    }
		  if (ppl == pickup.end())
		    {
		      ++p;
		      continue;
		    }
		}
	    }

	  pickup.insert(piwl.front().id());
	  capacity -= piwl.front().weight();

	  p = piwl.erase(p);
	}
    }

  if (! pickup.empty())
    { acts.adjust_pickup(pickup, value_pickup); }
}

Pickup_strategy pickup_strategy;

Add_strategy add_pickup_strategy(pickup_strategy);

// If we are carrying any packages, go to the delivery point of the
// heaviest package.

class
Delivery_strategy : public Strategy
{
  void operator()(State&, Package_list&, Actions&) const;
};

void
Delivery_strategy::operator()(
  State& state,
  Package_list&,
  Actions& acts) const
{
  Package_list mpl = state.me().packages(state.robots());
  if (mpl.empty())
    { return; }

  int heaviest = -1;
  int weight = -1;
  int dest_x = -1;
  int dest_y = -1;
  for (Package_list::const_iterator p = mpl.begin();
       p != mpl.end();
       ++p)
    {
      int pw = state.packs().weight(*p);
      int px;
      int py;
      if (pw >= 0 && pw > weight
	  && state.packs().dest(*p, px, py)
	  && (px != state.xme() || py != state.yme()))
	{
	  heaviest = *p;
	  weight = pw;
	  dest_x = px;
	  dest_y = py;
	}
    }

  if (heaviest >= 0)
    {
      acts.adjust_move_to(state.board(), state.me(), state.robots(),
			  dest_x, dest_y, value_move_to_drop);
    }
}

Delivery_strategy delivery_strategy;

Add_strategy add_delivery_strategy(delivery_strategy);

// If we are at the destination of any packages we are carrying, drop them.

class Drop_strategy : public Strategy
{
  void operator()(State&, Package_list&, Actions&) const;
};

void
Drop_strategy::operator()(
  State& state,
  Package_list&,
  Actions& acts) const
{
  Package_list mpl = state.me().packages(state.robots());
  if (mpl.empty())
    { return; }

  Package_list drop;
  for (Package_list::const_iterator p = mpl.begin();
       p != mpl.end();
       ++p)
    {
      int x;
      int y;
      if (state.packs().dest(*p, x, y)
	  && x == state.xme()
	  && y == state.yme())
	{ drop.insert(*p); }
    }

  if (! drop.empty())
    { acts.adjust_drop(drop, value_drop); }
}

Drop_strategy drop_strategy;

Add_strategy add_drop_strategy(drop_strategy);

// Try to avoid nearby robots.

class Avoid_robot_strategy : public Strategy
{
  void operator()(State&, Package_list&, Actions&) const;
};

void
Avoid_robot_strategy::operator()(
  State& state,
  Package_list&,
  Actions& acts) const
{
  State::Robot_list rl = state.nearby_robots();
  if (rl.empty())
    { return; }

  bool robot_dirs[direction_max];
  for (Direction d = direction_north; d < direction_max; ++d)
    { robot_dirs[d] = false; }

  for (State::Robot_list::const_iterator pr = rl.begin();
       pr != rl.end();
       ++pr)
    {
      bool this_robot_dirs[direction_max];
      state.board().direction(state.xme(), state.yme(), pr->x, pr->y,
			      this_robot_dirs);
      for (Direction d = direction_north; d < direction_max; ++d)
	{
	  if (this_robot_dirs[d])
	    { robot_dirs[d] = true; }
	}
    }

  for (Direction d = direction_north; d < direction_max; ++d)
    {
      if (robot_dirs[d])
	{ acts.adjust_move(d, value_meet_robot); }
    }
}

Avoid_robot_strategy avoid_robot_strategy;

Add_strategy add_avoid_robot_strategy(avoid_robot_strategy);

// Don't get caught between a robot and water.

class Dry_robot_strategy : public Strategy
{
  void operator()(State&, Package_list&, Actions&) const;
};

void
Dry_robot_strategy::operator()(
  State& state,
  Package_list&,
  Actions& acts) const
{
  // Don't worry about this if we are not next to water.  A more
  // clever strategy would consider where we are going to be next
  // turn.
  bool lethal_dirs[direction_max];
  int lethal_count = state.board().lethal_direction(state.xme(),
						    state.yme(),
						    lethal_dirs);
  if (lethal_count == 0)
    { return; }

  State::Robot_list rl = state.nearby_robots();
  if (rl.empty())
    { return; }

  // Avoid moving to another location which is next to water.
  bool secondary_lethal_dirs[direction_max];
  for (Direction d = direction_north; d < direction_max; ++d)
    {
      secondary_lethal_dirs[d] = false;
      if (! lethal_dirs[d])
	{
	  int xd = state.xme();
	  int yd = state.yme();
	  in_direction(d, xd, yd);
	  bool dummy_dirs[direction_max];
	  if (state.board().lethal_direction(xd, yd, dummy_dirs) > 0)
	    { secondary_lethal_dirs[d] = true; }
	}
    }

  // There are any nearby robots and we are next to water.  Move away
  // from the water, but, if possible, not toward the other robots.

  bool robot_dirs[direction_max];
  for (Direction d = direction_north; d < direction_max; ++d)
    { robot_dirs[d] = false; }

  bool must_move = false;

  for (State::Robot_list::const_iterator pr = rl.begin();
       pr != rl.end();
       ++pr)
    {
      if (pr->distance <= 3)
	{
	  // If we are next to water, move away from it, but, if
	  // possible, not toward the other robot.

	  bool this_robot_dirs[direction_max];
	  state.board().direction(state.xme(), state.yme(), pr->x, pr->y,
				  this_robot_dirs);

	  for (Direction d = direction_north; d < direction_max; ++d)
	    {
	      if (this_robot_dirs[d])
		{ robot_dirs[d] = true; }
	    }

	  // If this robot is right next to us, and is in a position
	  // to push us into the water, then we must move.
	  if (pr->distance == 1)
	    {
	      for (Direction d = direction_north; d < direction_max; ++d)
		{
		  if (this_robot_dirs[d]
		      && lethal_dirs[reverse_direction(d)])
		    {
		      must_move = true;
		      break;
		    }
		}
	    }
	}
    }

  int value = value_dry_robot;
  if (must_move)
    { value = value_dry_endangered_robot; }

  bool any = false;
  for (Direction d = direction_north; d < direction_max; ++d)
    {
      if (! lethal_dirs[d] && ! secondary_lethal_dirs[d] && ! robot_dirs[d])
	{
	  any = true;
	  acts.adjust_move(d, value);
	}
    }

  // If we couldn't find any direction away from water which was not
  // toward a robot, then move toward a robot.
  if (! any)
    {
      bool any = false;
      for (Direction d = direction_north; d < direction_max; ++d)
	{
	  if (! lethal_dirs[d] && ! secondary_lethal_dirs[d])
	    {
	      any = true;
	      acts.adjust_move(d, value);
	    }
	}
    }

  // If our only good moves still leave us next to water, then don't
  // worry about it.  A more sophisticated approach would avoid a
  // potential death square by checking whether another robot is in a
  // position to kill us as we move in one of the secondary_lethal_dirs.
}

Dry_robot_strategy dry_robot_strategy;

Add_strategy add_dry_robot_strategy(dry_robot_strategy);

// If we've been bumped a lot, do something different.

class Avoid_bump_strategy : public Strategy
{
  void operator()(State&, Package_list&, Actions&) const;
  int order() const
  { return 10; }
};

void
Avoid_bump_strategy::operator()(
  State& state,
  Package_list& pl,
  Actions& acts) const
{
  if (state.me().bumped_count() < 4)
    { return; }

  // We've been bumped a lot.  Try to do something different to break
  // the loop.  But if there is a package here, try to pick it up anyhow.
  if (! pl.empty())
    { return; }

  bool valid_dirs[direction_max];
  state.board().valid_direction(state.xme(), state.yme(), valid_dirs);

  Direction best = acts.best_move();
  if (best == direction_max)
    { best = direction_north; }

  Direction tryd = direction_right(best);
  if (! valid_dirs[tryd])
    {
      tryd = direction_right(reverse_direction(best));
      if (! valid_dirs[tryd])
	{
	  tryd = reverse_direction(best);
	  if (! valid_dirs[tryd])
	    { tryd = best; }
	}
    }

  int value = value_avoid_bump;
  if (state.me().bumped_count() > 8)
    { value = value_avoid_many_bumps; }

  if (valid_dirs[tryd])
    { acts.adjust_move(tryd, value_avoid_bump); }
}

Avoid_bump_strategy avoid_bump_strategy;

Add_strategy add_avoid_bump_strategy(avoid_bump_strategy);

// If we aren't going to do anything else, move at random.  It may be
// better than nothing.

class Random_strategy : public Strategy
{
  void operator()(State&, Package_list&, Actions&) const;
  int order() const
  { return 100; }
};

void
Random_strategy::operator()(
  State& state,
  Package_list&,
  Actions& acts) const
{
  Direction best = acts.best_move();
  if (best == direction_max)
    {
      bool valid_dirs[direction_max];
      state.board().valid_direction(state.xme(), state.yme(), valid_dirs);
      int c = 0;
      for (Direction d = direction_north; d < direction_max; ++d)
	{
	  if (valid_dirs[d])
	    { ++c; }
	}
      int i = random() % c;
      for (Direction d = direction_north; d < direction_max; ++d)
	{
	  if (valid_dirs[d])
	    {
	      if (i == 0)
		{
		  acts.adjust_move(d, value_random);
		  return;
		}
	      --i;
	    }
	}
    }
}

Random_strategy random_strategy;

Add_strategy add_random_strategy(random_strategy);

// If we happen to have a chance to kill another robot by pushing it
// into the water, opportunistically take it.

class Kill_strategy : public Strategy
{
  void operator()(State&, Package_list&, Actions&) const;
};

void
Kill_strategy::operator()(
  State& state,
  Package_list&,
  Actions& acts) const
{
  // We're going to bid for this, so don't do it if we are low on
  // money.
  if (state.me().money() < 50)
    { return; }

  State::Robot_list rl = state.nearby_robots();
  for (State::Robot_list::const_iterator p = rl.begin();
       p != rl.end();
       ++p)
    {
      if (p->distance == 1)
	{
	  bool dirs[direction_max];
	  state.board().direction(state.xme(), state.yme(), p->x, p->y,
				  dirs);
	  for (Direction d = direction_north; d < direction_max; ++d)
	    {
	      if (dirs[d])
		{
		  int x = p->x;
		  int y = p->y;
		  in_direction(d, x, y);
		  if (state.board().at(x, y) == Board::tile_lethal)
		    {
		      acts.adjust_move(d, value_kill_robot, time_first);
		      return;
		    }
		}
	    }
	}
    }
}

Kill_strategy kill_strategy;

Add_strategy add_kill_strategy(kill_strategy);
