// strategy.h  -*- C++ -*-
// Ian Lance Taylor for ICFP 2002

// Define strategies which make proposals for actions in a given state.

#include <queue>
#include <string>

#include "board.h"
#include "player.h"
#include "robots.h"
#include "package.h"
#include "state.h"

// The Actions class holds the different actions we could make in this
// turn.  Scores are set for these actions by the various strategies.

enum Time
  {
    time_first,
    time_last,
    time_any,
    time_max
  };

class Actions
{
public:
  Actions()
    : m_pickup(), m_drop()
  { memset(this->m_move, 0, sizeof this->m_move); }

  void adjust_move(Direction d, int adjust, Time time = time_any);
  void adjust_pickup(const Package_list& pl, int adjust, Time time = time_any);
  void adjust_drop(const Package_list& pl, int adjust, Time time = time_any);

  void adjust_move_to(Board&, const Player&, const Robots&, int x, int y,
		      int adjust, Time time = time_any);

  // Return the command to execute for this action, and set the price,
  // the expected location, and the package list.
  std::string command(State&, int& bid, int& xe, int& ye, Package_list& pe);

  bool any_drops() const
  { return ! this->m_drop.empty(); }

  Direction best_move() const;

private:
  struct Time_scores
  {
    int scores[time_max];
  };

  typedef std::map<Package_list, Time_scores> Package_score;

  void get_score(const int* pt, int& score, Time& time) const;

  int m_move[direction_max][time_max];
  Package_score m_pickup;
  Package_score m_drop;
};

class Strategy
{
public:
  virtual void operator()(State&, Package_list&, Actions&) const = 0;
  virtual int order() const
  { return 0; }
};

inline int operator<(const Strategy& s1, const Strategy& s2)
{ return s1.order() < s2.order(); }

typedef std::list<Strategy*> Strategy_list;

extern Strategy_list strategies;

// This class is used to build the strategy list at global constructor
// time.

class Add_strategy
{
public:
  Add_strategy(Strategy& strat)
  { strategies.push_back(&strat); }
};
